/*
 * Copyright (c) 1998, 2002 Sandro Sigala.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gamewidget.h"
#include "images.h"

FXDEFMAP(GameWidget) GameWidgetMap[] = {
  FXMAPFUNC(SEL_PAINT, 0, GameWidget::onPaint),
};

FXIMPLEMENT(GameWidget, FXCanvas, GameWidgetMap, ARRAYNUMBER(GameWidgetMap))

GameWidget::GameWidget()
{
}

GameWidget::GameWidget(FXComposite *p, FXObject *tgt, FXSelector sel, FXuint opts,
		       FXint x, FXint y, FXint w, FXint h)
	: FXCanvas(p, tgt, sel, opts, x, y, w, h)
{
	loadSprites();

	computeGameSize();

       	dbuf = new FXBMPImage(getApp(), 0, IMAGE_SHMI|IMAGE_SHMP, getWidth(), getHeight());
       	dbuf->create();
}

GameWidget::~GameWidget()
{
	for (int i = 0; i < num_sprites; ++i)
		delete sprites[i];
	for (int i = 0; i < num_masks; ++i)
		delete masks[i];

	delete dbuf;
}

void GameWidget::computeGameSize()
{
 	xb = (getWidth() % sprite_width) / 2;
	gwidth = getWidth() / sprite_width;
	yb = (getHeight() % sprite_height) / 2;
	gheight = getHeight() / sprite_height;
	scr_dwidth = gwidth / 2;
	scr_dheight = gheight / 2;
	if (scr_dwidth == 0)
		++scr_dwidth;
	if (scr_dheight == 0)
		++scr_dheight;
}

void GameWidget::loadSprites()
{
	loadMaskedSprite(Sprite_Zacman_Up, img_zacman_up, img_zacman_up_mask);
	loadMaskedSprite(Sprite_Zacman_Down, img_zacman_down, img_zacman_down_mask);
	loadMaskedSprite(Sprite_Zacman_Left, img_zacman_left, img_zacman_left_mask);
	loadMaskedSprite(Sprite_Zacman_Right, img_zacman_right, img_zacman_right_mask);
	loadMaskedSprite(Sprite_Zacman_Potion_Up, img_zacman_potion_up, img_zacman_up_mask);
	loadMaskedSprite(Sprite_Zacman_Potion_Down, img_zacman_potion_down, img_zacman_down_mask);
	loadMaskedSprite(Sprite_Zacman_Potion_Left, img_zacman_potion_left, img_zacman_left_mask);
	loadMaskedSprite(Sprite_Zacman_Potion_Right, img_zacman_potion_right, img_zacman_right_mask);
	loadMaskedSprite(Sprite_Zacman_Damaged_Up, img_zacman_damaged_up, img_zacman_up_mask);
	loadMaskedSprite(Sprite_Zacman_Damaged_Down, img_zacman_damaged_down, img_zacman_down_mask);
	loadMaskedSprite(Sprite_Zacman_Damaged_Left, img_zacman_damaged_left, img_zacman_left_mask);
	loadMaskedSprite(Sprite_Zacman_Damaged_Right, img_zacman_damaged_right, img_zacman_right_mask);
	loadMaskedSprite(Sprite_Monster_Up, img_monster_up, img_monster_up_mask);
	loadMaskedSprite(Sprite_Monster_Down, img_monster_down, img_monster_down_mask);
	loadMaskedSprite(Sprite_Monster_Left, img_monster_left, img_monster_left_mask);
	loadMaskedSprite(Sprite_Monster_Right, img_monster_right, img_monster_right_mask);
	loadMaskedSprite(Sprite_Pill, img_pill, img_pill_mask);
	loadMaskedSprite(Sprite_Potion, img_potion, img_potion_mask);
	loadMaskedSprite(Sprite_Arrow_Up, img_arrow_up, img_arrow_up_mask);
	loadMaskedSprite(Sprite_Arrow_Down, img_arrow_down, img_arrow_down_mask);
	loadMaskedSprite(Sprite_Balloon_Ouch, img_balloon_ouch, img_balloon_ouch_mask);
	loadMaskedSprite(Sprite_Balloon_Chomp, img_balloon_chomp, img_balloon_chomp_mask);
	loadMaskedSprite(Sprite_Balloon_Exit, img_balloon_exit, img_balloon_exit_mask);

	loadSprite(Sprite_Back1, img_back1);
	loadSprite(Sprite_Back2, img_back2);
	loadSprite(Sprite_Back3, img_back3);
	loadSprite(Sprite_Back4, img_back4);
	loadSprite(Sprite_Back5, img_back5);
	loadSprite(Sprite_Back6, img_back6);
	loadSprite(Sprite_Wall1, img_wall1);
	loadSprite(Sprite_Wall2, img_wall2);
	loadSprite(Sprite_Wall3, img_wall3);
	loadSprite(Sprite_Wall4, img_wall4);
	loadSprite(Sprite_Wall5, img_wall5);
	loadSprite(Sprite_Wall6, img_wall6);
	loadSprite(Sprite_Pause, img_pause);
}

void GameWidget::loadSprite(Sprite spr, const unsigned char *s)
{
	sprites[spr] = new FXBMPImage(getApp(), s, IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
	sprites[spr]->create();
}

void GameWidget::loadMaskedSprite(Sprite spr, const unsigned char *s1, const unsigned char *s2)
{
	sprites[spr] = new FXBMPImage(getApp(), s1, IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
	sprites[spr]->create();
	masks[spr] = new FXBMPImage(getApp(), s2, IMAGE_KEEP|IMAGE_SHMI|IMAGE_SHMP);
	masks[spr]->create();
}

void GameWidget::create()
{
	FXCanvas::create();
}

FXint GameWidget::getDefaultWidth()
{
	return gwidth;
}

FXint GameWidget::getDefaultHeight()
{
	return gheight;
}

void GameWidget::drawBoard()
{
 	if (last_width != getWidth() || last_height != getHeight()) {
		dbuf->resize(getWidth(), getHeight());
		computeGameSize();
		last_width = getWidth();
		last_height = getHeight();
	}

	FXDCWindow dc(dbuf);
	dc.setForeground(FXRGB(0, 0, 0));
	dc.fillRectangle(0, 0, getWidth(), getHeight());

	for (int y = 0; y < gheight; ++y)
		for (int x = 0; x < gwidth; ++x) {
			int mx = x + engine.xpos - gwidth / 2;
			int my = y + engine.ypos - gheight / 2;
			Sprite s = engine.xy(mx, my);
			if (s == Sprite_Pill || s == Sprite_Potion) { // masked sprites
 			   	dc.setFunction(BLT_SRC);
				dc.drawImage(sprites[engine.sprite_back], x*sprite_width + xb, y*sprite_height + yb);
			 	dc.setFunction(BLT_NOT_SRC_OR_DST);
				dc.drawImage(masks[s], x*sprite_width + xb, y*sprite_height + yb);
				dc.setFunction(BLT_SRC_AND_DST);
				dc.drawImage(sprites[s], x*sprite_width + xb, y*sprite_height + yb);
			} else {
   			   	dc.setFunction(BLT_SRC);
				dc.drawImage(sprites[s], x*sprite_width + xb, y*sprite_height + yb);
			}

			//  Enhance the 3d effect of walls
			if (s == engine.sprite_wall) {
				dc.setForeground(FXRGB(255, 255, 255));
				if (engine.xy(mx-1, my) != engine.sprite_wall)
					dc.drawLine(x*sprite_width + xb, y*sprite_height + yb,
						x*sprite_width + xb, (y+1)*sprite_height + yb);
				if (engine.xy(mx, my-1) != engine.sprite_wall)
					dc.drawLine(x*sprite_width + xb, y*sprite_height + yb,
						(x+1)*sprite_width + xb, y*sprite_height + yb);
				dc.setForeground(FXRGB(32, 32, 32));
				if (engine.xy(mx+1, my) != engine.sprite_wall)
					dc.drawLine((x+1)*sprite_width + xb - 1, y*sprite_height + yb,
						(x+1)*sprite_width + xb - 1, (y+1)*sprite_height + yb);
				if (engine.xy(mx, my+1) != engine.sprite_wall)
					dc.drawLine(x*sprite_width + xb, (y+1)*sprite_height + yb - 1,
						(x+1)*sprite_width + xb, (y+1)*sprite_height + yb - 1);
   			}

			Monster *m;
			if ((m = engine.findMonster(mx, my)) != 0) {
			 	dc.setFunction(BLT_NOT_SRC_OR_DST);
				dc.drawImage(masks[m->sprite], x*sprite_width + xb, y*sprite_height + yb);
				dc.setFunction(BLT_SRC_AND_DST);
				dc.drawImage(sprites[m->sprite], x*sprite_width + xb, y*sprite_height + yb);
			}
		}

	if (engine.xpos <= scr_dwidth && engine.ypos < scr_dheight) {
		dc.setFunction(BLT_NOT_SRC_OR_DST);
		dc.drawImage(masks[Sprite_Arrow_Down], (scr_dwidth - engine.xpos) * sprite_width + xb,
 		     	     (scr_dheight - engine.ypos - 1) * sprite_height + yb);
		dc.setFunction(BLT_SRC_AND_DST);
		dc.drawImage(sprites[Sprite_Arrow_Down], (scr_dwidth - engine.xpos) * sprite_width + xb,
 			     (scr_dheight - engine.ypos - 1) * sprite_height + yb);
	}
	if (engine.xpos >= engine.map_width - scr_dwidth - 1 && engine.ypos > engine.map_height - scr_dheight - 1) {
		dc.setFunction(BLT_NOT_SRC_OR_DST);
		dc.drawImage(masks[Sprite_Arrow_Up], (scr_dwidth - engine.xpos + engine.map_width - 1) * sprite_width + xb,
 		     	     (scr_dheight - engine.ypos + engine.map_height) * sprite_height + yb);
		dc.setFunction(BLT_SRC_AND_DST);
		dc.drawImage(sprites[Sprite_Arrow_Up], (scr_dwidth - engine.xpos + engine.map_width - 1) * sprite_width + xb,
 		     	     (scr_dheight - engine.ypos + engine.map_height) * sprite_height + yb);
	}

	Sprite s = Sprite_Zacman_Up; // Avoid uninitialized warning
 	if (engine.status == Status_Died)
 		switch (engine.dir) {
		case North: s = Sprite_Zacman_Damaged_Up; break;
		case South: s = Sprite_Zacman_Damaged_Down; break;
		case West: s = Sprite_Zacman_Damaged_Left; break;
		case East: s = Sprite_Zacman_Damaged_Right;
		}
        else if (engine.potion_time > 0)
		switch (engine.dir) {
		case North: s = Sprite_Zacman_Potion_Up; break;
		case South: s = Sprite_Zacman_Potion_Down; break;
		case West: s = Sprite_Zacman_Potion_Left; break;
		case East: s = Sprite_Zacman_Potion_Right;
		}
	else
		switch (engine.dir) {
		case North: s = Sprite_Zacman_Up; break;
		case South: s = Sprite_Zacman_Down; break;
		case West: s = Sprite_Zacman_Left; break;
		case East: s = Sprite_Zacman_Right;
		}

	dc.setFunction(BLT_NOT_SRC_OR_DST);
	dc.drawImage(masks[s], scr_dwidth * sprite_width + xb, scr_dheight * sprite_height + yb);
	dc.setFunction(BLT_SRC_AND_DST);
	dc.drawImage(sprites[s], scr_dwidth * sprite_width + xb, scr_dheight * sprite_height + yb);

	if (engine.balloon_time > 0) {
		dc.setFunction(BLT_NOT_SRC_OR_DST);
		dc.drawImage(masks[engine.balloon], scr_dwidth * sprite_width + xb - masks[engine.balloon]->getWidth(),
  				scr_dheight * sprite_height + yb - masks[engine.balloon]->getHeight());
		dc.setFunction(BLT_SRC_AND_DST);
		dc.drawImage(sprites[engine.balloon], scr_dwidth * sprite_width + xb - masks[engine.balloon]->getWidth(),
  				scr_dheight * sprite_height + yb - masks[engine.balloon]->getHeight());
	}

	if (engine.paused) {
		dc.setFunction(BLT_SRC);
		dc.drawImage(sprites[Sprite_Pause], (getWidth() - sprites[Sprite_Pause]->getWidth()) / 2,
			     (getHeight() - sprites[Sprite_Pause]->getHeight()) / 2);
 	}
}

long GameWidget::onPaint(FXObject *, FXSelector, void *ptr)
{
	drawBoard();
 	FXEvent *ev = (FXEvent *)ptr;
	FXDCWindow dc(this, ev);
  	dc.drawImage(dbuf, 0, 0);
  	return 1;
}

